#include <stdio.h>
#include <unistd.h>
#include <fcntl.h>
#include <stdlib.h>
#include <termios.h>
#include <sys/mman.h>
#include <sys/time.h>
#include <sys/types.h>
#include "hwlib.h"
#include "socal/socal.h"
#include "socal/hps.h"
#include "socal/alt_gpio.h"
#include "hps_0.h"

#define HW_REGS_BASE ( ALT_STM_OFST )
#define HW_REGS_SPAN ( 0x04000000 )
#define HW_REGS_MASK ( HW_REGS_SPAN - 1 )

int hextobin(void);
void comtrig(void);
int comread(unsigned int *d, unsigned int *c);

void *h2p_lw_sw_addr;
void *h2p_lw_pin0_addr;
void *h2p_lw_pout0_addr;
void *h2p_lw_pin1_addr;
void *h2p_lw_pout1_addr;
void *h2p_lw_pin2_addr;
void *h2p_lw_pout2_addr;
void *virtual_base;

char          recbuf[256];                       // ^[~i(M)obt@
unsigned int  cdata, comd;                       // R}hf[^, R}hڍ׃f[^
unsigned int  comport2on  = 0x87654321;          // R}h݃gKON
unsigned int  comport2off = 0x12345678;          // R}h݃gKOFF
unsigned int  outstate = 0, exoutstate = 0;      // GPIOo͏
unsigned int  inpolarity = 0, exinpolarity = 0;  // GPIO͋ɐ 0:normal 1:reverse
unsigned int  outpolarity = 0, exoutpolarity = 0; // GPIOo͓͋ɐ 0:normal 1:reverse
unsigned int  arduinostate = 0;                  // ARDUINOo͏
unsigned int  arduinodir = 0;                    // ARDUINOo͕
unsigned int  countmode[6] = {0,0,0,0,0,0};      // JE^샂[h
unsigned int  countlast[6] = {0xffffffff,0xffffffff,0xffffffff,0xffffffff,0xffffffff,0xffffffff}; // JE^ŏIl
unsigned int  countoffset[6] = {0,0,0,0,0,0};    // JE^l
unsigned int  countfilter[6] = {0,0,0,0,0,0};    // JE^ptB^l
unsigned int  countoutdis = 0;                   // JE^o͂Ȃǖ(bit0)APWMo͂ɕύX(bit1)
unsigned int  pstate[8]   = {0,0,0,0,0,0,0,0};   // [VRg[ e
unsigned int  ppuls[8]    = {0,0,0,0,0,0,0,0};   // [VRg[ e݈ړ 6 ^C}[ݒl 2
unsigned int  pposi[8]    = {0,0,0,0,0,0,0,0};   // [VRg[ e݈ʒu
unsigned int  poffset[8]  = {0,0,0,0,0,0,0,0};   // [VRg[ el
unsigned int  movedir[8]  = {0,0,0,0,0,0,0,0};   // [VRg[ eړ
unsigned int  movepuls[8] = {0,0,0,0,0,0,0,0};   // [VRg[ ewړ 6 ^C}[l 2
unsigned int  movevelo[8] = {1000,1000,1000,1000,1000,1000,1000,1000};   // [VRg[ ew葬x
unsigned int  moveacce[8] = {100,100,100,100,100,100,100,100};   // [VRg[ ewx
unsigned int  masteraxis  = 0;                   // }X^[ԍ
unsigned int  pmaster     = 0;                   // }X^[ bit5:5`bit0:0
unsigned int  pslave      = 0;                   // X[u bit5:5`bit0:0
unsigned int  mspattern   = 0;                   // Ԏw莲
unsigned int  limitsw     = 0;                   // ~bgXCb`ݒ
unsigned int  poutmode    = 0;                   // pXo̓[h

unsigned int  adcstart = 0;                      // ADC X^[g
unsigned int  adccycle = 0;                      // ADC XVTCN
unsigned int  adcchan = 0;                       // ADC `lۑ
unsigned int  adcinterval = 10000;               // ADC C^[o unit = 0.1us
unsigned int  adcsd = 0;                         // ADC /VO :1
unsigned int  adcmode = 0;                       // ADC 샂[h
unsigned int  adcdata = 0;                       // ADC f[^

unsigned int  pwmwidth[16] = {25000,25000,25000,25000,25000,25000,25000,25000,
                              25000,25000,25000,25000,25000,25000,0,500000};
                                                 // PWMpXA15Ԗڂstrat/stopA16ԖڂPWMTCN
int oldf;

int main() {

    int fd;
    int i, p, q, r, s, t, cntno, axisno;
    unsigned int u;
    char swID;

    // FPGA Device/Memory OPEN

    if( ( fd = open( "/dev/mem", ( O_RDWR | O_SYNC ) ) ) == -1 ) {
        printf( "OPEN_ERR%c",0xd);
        return( 1 );
    }

    // FPGA Device/Memory x[XAhX擾
    
    virtual_base = mmap( NULL, HW_REGS_SPAN, ( PROT_READ | PROT_WRITE ), MAP_SHARED, fd, HW_REGS_BASE );

    if( virtual_base == MAP_FAILED ) {
        printf( "MMAP_ERR%c",0xd);
        close( fd );
        return( 1 );
    }
    
    // FPGA |[gAhXZo
    
    h2p_lw_sw_addr=virtual_base + ( ( unsigned long  )( ALT_LWFPGASLVS_OFST + PIO_SW_BASE ) & ( unsigned long)( HW_REGS_MASK ) );
    h2p_lw_pin0_addr=virtual_base + ( ( unsigned long  )( ALT_LWFPGASLVS_OFST + PIO_IN0_BASE ) & ( unsigned long)( HW_REGS_MASK ) );
    h2p_lw_pout0_addr=virtual_base + ( ( unsigned long  )( ALT_LWFPGASLVS_OFST + PIO_OUT0_BASE ) & ( unsigned long)( HW_REGS_MASK ) );
    h2p_lw_pin1_addr=virtual_base + ( ( unsigned long  )( ALT_LWFPGASLVS_OFST + PIO_IN1_BASE ) & ( unsigned long)( HW_REGS_MASK ) );
    h2p_lw_pout1_addr=virtual_base + ( ( unsigned long  )( ALT_LWFPGASLVS_OFST + PIO_OUT1_BASE ) & ( unsigned long)( HW_REGS_MASK ) );
    h2p_lw_pin2_addr=virtual_base + ( ( unsigned long  )( ALT_LWFPGASLVS_OFST + PIO_IN2_BASE ) & ( unsigned long)( HW_REGS_MASK ) );
    h2p_lw_pout2_addr=virtual_base + ( ( unsigned long  )( ALT_LWFPGASLVS_OFST + PIO_OUT2_BASE ) & ( unsigned long)( HW_REGS_MASK ) );

    // XChXCb`̐ݒ擾
    
    swID = (*(uint32_t *)h2p_lw_sw_addr) & 0xf;
    if(swID == 0xf){                            // ÎƂ
        close(fd);
        return(1);
    }
    
    // GPIOo,JE^ݒ

    *(uint32_t *)h2p_lw_pout0_addr = 0;
    *(uint32_t *)h2p_lw_pout1_addr = 0;
    *(uint32_t *)h2p_lw_pout2_addr = comport2off;

    *(uint32_t *)h2p_lw_pout1_addr = 0x110000;  // X^[gAbvL[R[hǎw|[g1ɃZbg
    comtrig();                                  // R}h݃gK
    comread(&comd, &cdata);                     // ǎs
    
    if(cdata != 0xDACE350){                     // ݒ̂Ƃ
        *(uint32_t *)h2p_lw_pout0_addr = 0xDACE350;         // X^[gAbvL[R[h|[g0ɃZbg
        *(uint32_t *)h2p_lw_pout1_addr = 0x100000;          // X^[gAbvL[R[hw|[g1ɃZbg
        comtrig();                                          // R}h݃gK
        for(i=0; i!= 6; i++){
            p = i << 16;
            *(uint32_t *)h2p_lw_pout0_addr = movevelo[i];   // x|[g0ɃZbg
            *(uint32_t *)h2p_lw_pout1_addr = p | 0x2100000; // xݒ |[g1f[^
            comtrig();                                      // R}h݃gK
            *(uint32_t *)h2p_lw_pout0_addr = moveacce[i];   // x|[g0ɃZbg
            *(uint32_t *)h2p_lw_pout1_addr = p | 0x2200000; // xݒ |[g1f[^
            comtrig();                                      // R}h݃gK
        }
    }
    else{                                       // ݒێ̂Ƃ
        *(uint32_t *)h2p_lw_pout1_addr = 0x20000;           // GPIOo͏ԓǎw
        comtrig();                                          // R}h݃gK
        comread(&exoutstate, &outstate);                    // ǎs
        exoutstate = exoutstate & 0xf;

        *(uint32_t *)h2p_lw_pout1_addr = 0x40000;           // GPIO͋ɐݒǎw
        comtrig();                                          // R}h݃gK
        comread(&exinpolarity, &inpolarity);                // ǎs
        exinpolarity = exinpolarity & 0xf;
        
        *(uint32_t *)h2p_lw_pout1_addr = 0x60000;           // GPIOo͋ɐݒǎw
        comtrig();                                          // R}h݃gK
        comread(&exoutpolarity, &outpolarity);              // ǎs
        exoutpolarity = exoutpolarity & 0xf;
        
        for(i=0; i!= 6; i++){                               // JE^ݒԓǎ
            *(uint32_t *)h2p_lw_pout1_addr = 0x1100000 | (i << 16);       // JE^ݒԓǎw
            comtrig();                                                    // R}h݃gK
            comread(&countmode[i], &countlast[i]);                        // ǎs
            countmode[i] = countmode[i] & 0x5f;
            
            *(uint32_t *)h2p_lw_pout1_addr = 0x1100000 | ((i+8) << 16);   // tB^lǎw
            comtrig();                                                    // R}h݃gK
            comread(&comd, &countfilter[i]);                              // ǎs
            
            *(uint32_t *)h2p_lw_pout1_addr = 0x1400000 | (i << 16);       // lǎw
            comtrig();                                                    // R}h݃gK
            comread(&comd, &countoffset[i]);                              // ǎs
        }
        for(i=0; i!= 6; i++){                               // [Vݒԓǎ
            *(uint32_t *)h2p_lw_pout1_addr = 0x2A00000 | (i << 16);       // lǎw
            comtrig();                                                    // R}h݃gK
            comread(&comd, &poffset[i]);                                  // ǎs
        }
    }
    swID = (swID & 7) + 0x30;                               // IDԍݒ
    
    oldf = fcntl(STDIN_FILENO, F_GETFL, 0);                 // Wo͂ NON-Block Ƃ
    fcntl(STDIN_FILENO, F_SETFL, oldf | O_NONBLOCK);

    /* W(^[~i)̃f[^ǎAFPGÃ|[gɃR}h       */
    /* FPGÃ|[gf[^ǎAWo(^[~i)ɃX|Xo͂ */
    /* AX^XN(*)R[h͂ɂďI                                */
    
    do{
        // ADCf[^̑M

        if(adcstart){                                           // ADC 쒆 ?
        
            u = *(uint32_t *)h2p_lw_pin2_addr & 0x30000000;
            
            if((u == 0x10000000) || (u == 0x20000000)){         // ADC 쒆̂Ƃ
                u = u >> 28;
                if((adccycle != 0) && (u != adccycle)){         // ŏ̃f[^ł͂Ȃf[^XVꂽƂ
                    *(uint32_t *)h2p_lw_pout1_addr = 0x3100000; // ADCf[^ǎR}hR[h|[g1ɃZbg
                    comtrig();                                  // R}h݃gK
                    i = comread(&comd, &cdata);                 // ǎs
                    if(i > 50) printf("FPGA_ERR   %c",0xd);     // BUSYOFFƂȂȂƂ̓G[
                    else{
                        comd = comd & 0xf;
                        if(u == 2){
                             adcdata = cdata << 20;             // 1f[^ۑ
                             adcchan = comd;                    // ADC`lۑ
                        }
                        else{
                            adcdata = adcdata | (cdata << 4);   // 2f[^1f[^ƍ
                            adcchan = adcchan - comd;           // 2f[^1f[^̃`lZo
                            if(adcsd){
                                if((adcchan = -1) || (adcchan = 0)) printf("v%c%1X%08X%c",swID,comd,adcdata,0xd);
                            }                                   // ADCf[^(쓮)𑗐M
                            else{
                                if((adcchan = -1) || (adcchan = 0)) printf("V%c%1X%08X%c",swID,comd,adcdata,0xd);
                            }                                   // ADCf[^(VOGh)𑗐M
                        }
                        adccycle = u;                           // ADC XVTCNZbg
                    }
                }
                if(adccycle == 0) adccycle = u;                 // ŏ̃f[^̂Ƃ́AADC XVTCNύX̂
            }
        }
        else adccycle = 0;                                      // ADC XVTCNZbg

        // 1s擾

        recbuf[0] = 0;                                      // R}hR[h
        recbuf[1] = 0;                                      // IDR[h

        if(fgets(recbuf,256,stdin) == NULL) continue;       // W(^[~i)當f[^
        
        // vOÎƂ

        if(recbuf[0] == '*'){
            printf("*%c",0xd);
            break;
        }

        // WR}h (fW^o) ̂Ƃ

        if((recbuf[0] == 'W') && (recbuf[1] == swID)){
            
            if((recbuf[2] == 'R') || (recbuf[2] == 'r')){            // o̓f[^ǎŵƂ
                printf("W%c%1X%08X%c",recbuf[1],exoutstate,outstate,0xd);
                continue;
            }
            cdata = outstate;                                        // ʏWR}ĥƂ
            comd  = exoutstate;
            if(hextobin() != 0) continue;                            // 16i̕oCi
            outstate   = cdata;
            exoutstate = comd;

            *(uint32_t *)h2p_lw_pout0_addr = outstate;               // GPIOobit31`0̃f[^|[g0ɃZbg
            *(uint32_t *)h2p_lw_pout1_addr = exoutstate | 0x10000;   // GPIOobit35`32WR}hR[h|[g1ɃZbg
            comtrig();                                               // R}h݃gK
            i = comread(&comd, &cdata);                              // ǎs
            comd = comd & 0xf;
            if(i > 50) printf("FPGA_ERR   %c",0xd);                  // BUSYOFFƂȂȂƂ̓G[
            else       printf("R%c%1X%08X%c",recbuf[1],comd,cdata,0xd); // Wo(^[~i)֕f[^(X|X)𑗐M
            continue;
        }
        
        // YR}h (fW^͋ɐݒ) ̂Ƃ

        if((recbuf[0] == 'Y') && (recbuf[1] == swID)){
            
            if((recbuf[2] == 'R') || (recbuf[2] == 'r')){            // ͋ɐݒǎŵƂ
                printf("Y%c%1X%08X%c",recbuf[1],exinpolarity,inpolarity,0xd);
                continue;
            }
            cdata = inpolarity;                                      // ʏYR}ĥƂ
            comd  = exinpolarity;
            if(hextobin() != 0) continue;                            // 16i̕oCi
            inpolarity   = cdata;
            exinpolarity = comd;

            *(uint32_t *)h2p_lw_pout0_addr = inpolarity;             // GPIO͋ɐbit31`0̃f[^|[g0ɃZbg
            *(uint32_t *)h2p_lw_pout1_addr = exinpolarity | 0x30000; // GPIO͋ɐbit35`32YR}hR[h|[g1ɃZbg
            comtrig();                                               // R}h݃gK

            printf("U%c%1X%08X%c",recbuf[1],exinpolarity,inpolarity,0xd);   // 𑗐M
            continue;
        }
        
        // yR}h (fW^o͋ɐݒ) ̂Ƃ

        if((recbuf[0] == 'y') && (recbuf[1] == swID)){
            
            if((recbuf[2] == 'R') || (recbuf[2] == 'r')){            // o͋ɐݒǎŵƂ
                printf("y%c%1X%08X%c",recbuf[1],exoutpolarity,outpolarity,0xd);
                continue;
            }
            cdata = outpolarity;                                     // ʏyR}ĥƂ
            comd  = exoutpolarity;
            if(hextobin() != 0) continue;                            // 16i̕oCi
            outpolarity   = cdata;
            exoutpolarity = comd;

            *(uint32_t *)h2p_lw_pout0_addr = outpolarity;             // GPIOo͋ɐbit31`0̃f[^|[g0ɃZbg
            *(uint32_t *)h2p_lw_pout1_addr = exoutpolarity | 0x50000; // GPIOo͋ɐbit35`32yR}hR[h|[g1ɃZbg
            comtrig();                                               // R}h݃gK

            printf("U%c%1X%08X%c",recbuf[1],exoutpolarity,outpolarity,0xd);   // 𑗐M
            continue;
        }
        
        // wR}h (ARDUINOo) ̂Ƃ

        if((recbuf[0] == 'w') && (recbuf[1] == swID)){
            
            if((recbuf[2] == 'R') || (recbuf[2] == 'r')){            // o̓f[^ǎŵƂ
                printf("w%c0%08X%c",recbuf[1],arduinostate,0xd);
                continue;
            }
            cdata = arduinostate;                                    // ʏwR}ĥƂ
            comd  = 0;
            if(hextobin() != 0) continue;                            // 16i̕oCi
            arduinostate = cdata;

            cdata = cdata << 3;
            *(uint32_t *)h2p_lw_pout0_addr = cdata;                  // ARDUINOobit15`0̃f[^|[g0ɃZbg
            *(uint32_t *)h2p_lw_pout1_addr = 0x210000;               // wR}hR[h|[g1ɃZbg
            comtrig();                                               // R}h݃gK
            i = comread(&comd, &cdata);                              // ǎs
            cdata = cdata >> 3;
            if(i > 50) printf("FPGA_ERR   %c",0xd);                  // BUSYOFFƂȂȂƂ̓G[
            else       printf("r%c0%08X%c",recbuf[1],cdata,0xd);     // Wo(^[~i)֕f[^(X|X)𑗐M
            continue;
        }
        
        // zR}h (ARDUINOo͕w) ̂Ƃ

        if((recbuf[0] == 'z') && (recbuf[1] == swID)){
            
            if((recbuf[2] == 'R') || (recbuf[2] == 'r')){            // o͕ǎŵƂ
                printf("z%c0%08X%c",recbuf[1],arduinodir,0xd);
                continue;
            }
            cdata = arduinodir;                                      // ʏzR}ĥƂ
            comd  = 0;
            if(hextobin() != 0) continue;                            // 16i̕oCi
            arduinodir = cdata;

            cdata = cdata << 3;
            *(uint32_t *)h2p_lw_pout0_addr = cdata;                  // ARDUINOo͕bit15`0̃f[^|[g0ɃZbg
            *(uint32_t *)h2p_lw_pout1_addr = 0x220000;               // zR}hR[h|[g1ɃZbg
            comtrig();                                               // R}h݃gK

            printf("U%c0%08X%c",recbuf[1],arduinodir,0xd);           // 𑗐M
            continue;
        }

        // MR}h (JE^ݒ/JEglǎ) ̂Ƃ

        if((recbuf[0] == 'M') && (recbuf[1] == swID)){

            if((recbuf[2] == 'I') || (recbuf[2] == 'i') || (recbuf[2] == 'P') || (recbuf[2] == 'p')){
                                                    // o͂ƃJE^eXgpo͂𖳌APWMo͂ɕύX
                if(recbuf[2] == 'I') countoutdis = countoutdis | 1;          // o͂ƃJE^eXgpo͂𖳌
                else{
                    if(recbuf[2] == 'P') countoutdis = countoutdis | 2;      // PWMo͂ɕύX
                    else{
                        if(recbuf[2] == 'i') countoutdis = countoutdis & 2;  // o͂ƃJE^eXgpo͂Lɖ߂
                        else                 countoutdis = countoutdis & 1;  // PWMo͂̐ݒɖ߂
                    }
                }
                *(uint32_t *)h2p_lw_pout0_addr = countoutdis; // PWMo͂ɕύXbit|[g0ɃZbg
                *(uint32_t *)h2p_lw_pout1_addr = 0x1200000;   // ݒR[h|[g1ɃZbg
                comtrig();                          // R}h݃gK
                printf("M%c%c00000000%c",recbuf[1],recbuf[2],0xd);
                continue;
            }
            cdata = 0xFFFFFFFF;
            comd  = 0xFFFFFFFF;
            if(hextobin() != 0) continue;           // 16i̕oCi
            if(comd > 0x10) continue;               // JE^ԍŵƂ
            if(comd >= 8) cntno = comd -8;          // JE^ԍ擾
            else          cntno = comd;
            if(cntno > 5) continue;                 // JE^ԍŝƂ

            if((recbuf[3] == 'R') || (recbuf[3] == 'r')){ // JE^샂[hǎ̂Ƃ𑗐M
                if(comd < 8) printf("M%c%c%X0000000%c",recbuf[1],recbuf[2],((countmode[cntno] & 0x50) >> 4),0xd);
                else         printf("M%c%c%X%X000000%c",recbuf[1],recbuf[2],(countmode[cntno] & 0xf),((countmode[cntno] & 0x80) >> 4),0xd);
                continue;
            }
            i = countmode[cntno];                   // ݂̃JE^샂[h
            if(cdata != 0xFFFFFFFF){
                if(comd < 8){
                    if(cdata & 0x80000000) i = i | 0x10;    // JE^ JEgJn
                    if(cdata & 0x40000000) i = i & 0xcf;    // JE^ JEg~
                    if(cdata & 0x20000000) i = i | 0x40;    // JE^ ZbgfW^͖
                    else                   i = i & 0x9f;
                    if(cdata & 0x10000000) i = i | 0x20;    // JE^ ZbgON (̌AOFFƂKv͂Ȃ)
                }
                else{
                    i = (i & 0x50) | ((cdata >> 28) & 0xf); // JE^샂[h
                    p = (cdata >> 20) & 0xf0;
                    if(p != 0xf0) i = i | (p & 0x80);       // JE^샂[h(CW/CCW)
                }
            }
            if(comd < 8) *(uint32_t *)h2p_lw_pout0_addr = countlast[cntno];   // JEgŏIl|[g0ɃZbg
            else         *(uint32_t *)h2p_lw_pout0_addr = countfilter[cntno]; // tB^l|[g0ɃZbg
            countmode[cntno] = i & 0xdf;                    // JE^샂[hۑ (ZbgOFF)
            *(uint32_t *)h2p_lw_pout1_addr = (comd << 16) | 0x1000000 | i;    // MR}hR[h|[g1ɃZbg
            comtrig();                                      // R}h݃gK
            i = comread(&comd, &cdata);                     // ǎs
            if(i > 50) printf("FPGA_ERR   %c",0xd);         // BUSYOFFƂȂȂƂ̓G[
            else printf("N%c%c%08X%c",recbuf[1],recbuf[2],cdata,0xd); // Wo(^[~i)֕f[^(X|X)𑗐M
            continue;
        }
        
        // LR}h (JEgŏIl/lw) ̂Ƃ

        if((recbuf[0] == 'L') && (recbuf[1] == swID)){

            comd  = 0xFFFFFFFF;
            if(hextobin() != 0) continue;           // 16i̕oCi
                                                    // wf[^΍
            if(comd < 6){                           // ŏIlŵƂ
                cntno = comd;
                cdata = countlast[cntno];
            }
            else{
                if((comd > 7) && (comd < 14)){      // lŵƂ
                    cntno = comd - 8;
                    cdata = countoffset[cntno];
                }
                else continue;                      // JE^ԍŝƂ
            }
            if(hextobin() != 0) continue;           // ēxA16i̕oCi

            if((recbuf[3] == 'R') || (recbuf[3] == 'r')){  // JEgŏIl/lǍ݂̂Ƃ𑗐M
                printf("L%c%c%08X%c",recbuf[1],recbuf[2],cdata,0xd);
                continue;
            }
            if(comd < 6){
                countlast[cntno] = cdata;            // JEgŏIlۑ
                i = ((cntno << 16) | 0x1000000) | countmode[cntno]; // JEglǎ |[g1f[^
            }
            else{
                countoffset[cntno] = cdata;          // JEglۑ
                i = ((cntno << 16) | 0x1300000) | countmode[cntno]; // JEgl |[g1f[^
            }
            *(uint32_t *)h2p_lw_pout0_addr = cdata;  // JEgŏIl/l|[g0ɃZbg
            *(uint32_t *)h2p_lw_pout1_addr = i;      // MR}hR[h|[g1ɃZbg
            comtrig();                               // R}h݃gK
            
            printf("U%c%c%08X%c",recbuf[1],recbuf[2],cdata,0xd);   // 𑗐M
            continue;
        }

        // TR}h (JE^ptB^ݒ) ̂Ƃ

        if((recbuf[0] == 'T') && (recbuf[1] == swID)){

            cdata = 0xFFFFFFFF;
            comd  = 0xFFFFFFFF;
            if(hextobin() != 0) continue;           // 16i̕oCi
            if(comd > 5) continue;                  // JE^ԍŝƂ
            cntno = comd;
            if(cdata == 0xFFFFFFFF) cdata = countfilter[cntno]; // tB^lŵƂ

            if((recbuf[3] == 'R') || (recbuf[3] == 'r')){  // tB^lǍ݂̂Ƃ𑗐M
                printf("T%c%c%08X%c",recbuf[1],recbuf[2],countfilter[cntno],0xd);
                continue;
            }
            if(cdata > 0xfffff) cdata = 0xfffff;    // őlɐ
            countfilter[cntno] = cdata;             // tB^lۑ
            comd = comd + 8;
            i = ((comd << 16) | 0x1000000) | countmode[cntno]; // tB^l |[g1f[^

            *(uint32_t *)h2p_lw_pout0_addr = cdata; // tB^l|[g0ɃZbg
            *(uint32_t *)h2p_lw_pout1_addr = i;     // MR}hR[h|[g1ɃZbg
            comtrig();                              // R}h݃gK
            printf("U%c%c%08X%c",recbuf[1],recbuf[2],countfilter[cntno],0xd);   // 𑗐M
            continue;
        }

        // PR}h ([VRg[ړʐݒ) ̂Ƃ

        if((recbuf[0] == 'P') && (recbuf[1] == swID)){
            
            if((recbuf[2] == 'G') || (recbuf[2] == 'g')){       // ԈړJn̂Ƃ
                p = 0;                              // Ώ̎rbgp^[ݒ
                q = 1;
                for(i=3; i!= 9; i++){               // Ώ̎̌
                    if(recbuf[i] == '1') p = p | q;
                    else{
                        if(recbuf[i] != '0') break;
                    }
                    q = (q << 1);
                }
                if(i == 3) p = mspattern;           // w肪ȂƂ
                mspattern = p;                      // Ԏw莲ۑ
                if(p == 0) continue;                // w莲SȂƂ͖

                r = 0;                              // }X^[rbgp^[ݒ
                s = 0;                              // őړʏݒ
                q = 1;
                for(i= 0; i!= 6; i++){              // őړʂ̎T
                    if((p & q) && (movepuls[i] > s)){
                        r = q;                      // őړʎ̃rbgp^[XV
                        s = movepuls[i];            // őړʍXV
                        t = i;                      // őړʂ̎ԍXV
                    }
                    q = (q << 1);
                }
                if(s == 0) continue;                // S̈ړʂ0̂Ƃ͖
                masteraxis = t;                     // }X^[ԍۑ
                pmaster = r;                        // }X^[rbgp^[ bit5:5`bit0:0
                pslave  = p & ~r;                   // X[urbgp^[ bit5:5`bit0:0
                *(uint32_t *)h2p_lw_pout1_addr = 0x2300000 | (pslave << 8) | pmaster;
                                                    // }X^[X[uݒR[h|[g1ɃZbg
                comtrig();                          // R}h݃gK
                cdata = 0x80000000;                 // }X^[̈ړʂɖwƂ
                comd  = masteraxis + 8;             // }X^[ړJnƂ
            }
            else{
                if((recbuf[2] == 'S') || (recbuf[2] == 's')){   // Ԉړ~̂Ƃ
                    cdata = 0x80000000;             // }X^[̈ړʂɖwƂ
                    comd  = masteraxis;             // }X^[~w
                }
                else{                   // PƈړJn/~Aړʐݒ̂
                    cdata = 0x80000000;
                    comd  = 0xFFFFFFFF;
                    if(hextobin() != 0) continue;   // 16i̕oCi
                    if((recbuf[2] == 'L') || (recbuf[2] == 'l')){  // ~bgXCb`ݒ̂Ƃ
                        if(cdata == 0x80000000) cdata = limitsw;   // ŵƂ
                        limitsw = cdata & 0xffff;   // ~bgXCb`ݒۑ
                        *(uint32_t *)h2p_lw_pout1_addr = 0x2340000 | limitsw;
                                                    // ~bgXCb`ݒR[h|[g1ɃZbg
                        comtrig();                  // R}h݃gK

                        printf("U%c%c%08X%c",recbuf[1],recbuf[2],limitsw,0xd);   // 𑗐M
                        continue; 
                    }
                    if((recbuf[2] == 'M') || (recbuf[2] == 'm')){  // CW/CCW [hݒ̂Ƃ
                        if(cdata == 0x80000000) cdata = poutmode;  // ŵƂ
                        poutmode = cdata & 0x3f;    // pXo̓[hݒۑ
                        *(uint32_t *)h2p_lw_pout1_addr = 0x2380000 | poutmode;
                                                    // CW/CCW [hݒR[h|[g1ɃZbg
                        comtrig();                  // R}h݃gK

                        printf("U%c%c%08X%c",recbuf[1],recbuf[2],poutmode,0xd);   // 𑗐M
                        continue; 
                    }
                                                   // PƈړJn/~Aړʐݒ̂
                    if(comd == 0xFFFFFFFF) continue;   // ԍŝƂ
                }
            }
            if(comd >= 8) axisno = comd -8;
            else          axisno = comd;
            if(axisno > 7)  continue;               // ԍŝƂ
            *(uint32_t *)h2p_lw_pout1_addr = ((axisno << 16) | 0x2800000);
                                                    // [VRg[ԓǎR[h|[g1ɃZbg
            comtrig();                              // R}h݃gK
            i = comread(&pstate[axisno], &ppuls[axisno]);   // ǎs
            if(i > 50){
                printf("FPGA_ERR   %c",0xd);        // BUSYOFFƂȂȂƂ̓G[
                continue;
            }
            if(axisno < 6){
                if(((pstate[axisno] & 0x100) == 0) && (cdata != 0x80000000)){ // ړłȂړʖwłȂƂ
                    if(cdata & 0x80000000){
                        cdata = 0 - cdata;          // l(Βl)Ƃ
                        movedir[axisno] = 1;        // ړ ݒ
                    }
                    else  movedir[axisno] = 0;      // ړ ݒ
                    movepuls[axisno] = cdata;       // ړʕۑ
                }
            }
            else{
                 if(cdata != 0x80000000) movepuls[axisno] = cdata; // ^C}[lۑ
            }
            *(uint32_t *)h2p_lw_pout0_addr = movepuls[axisno];  // ړʂ|[g0ɃZbg
            i = ((axisno << 16) | 0x2000000) | movedir[axisno]; // ړJn ړʏ |[g1f[^
            if(((pstate[axisno] & 0x100) == 0) || (comd < 8)){
                                                    // ړłȂƂɂނ̂́A
                                                    //     ςă~bgXCb`Ƃ̂
                *(uint32_t *)h2p_lw_pout1_addr = i; // PR}hR[h|[g1ɃZbg
                comtrig();                          // R}h݃gK
            }
            if(((pstate[axisno] & 0x100) == 0) && (comd >= 8)){     // ړJnŵƂ
                i = i | 0x100;                      // ړJn ړʏ |[g1f[^
                *(uint32_t *)h2p_lw_pout1_addr = i; // PR}hR[h|[g1ɃZbg
                comtrig();                          // R}h݃gK
            }
            cdata = movepuls[axisno];
            if(movedir[axisno] == 1) cdata = 0 - cdata;        // ̂ƂlƂ
            printf("U%c%1X%08X%c",recbuf[1],comd,cdata,0xd);   // 𑗐M
            continue;
        }

        // VR}h ([VRg[x/ݒ) ̂Ƃ

        if((recbuf[0] == 'V') && (recbuf[1] == swID)){

            cdata = 0xFFFFFFFF;
            comd  = 0xFFFFFFFF;
            if(hextobin() != 0) continue;           // 16i̕oCi
            if(comd == 0xFFFFFFFF) continue;        // ԍŝƂ
            if(comd >= 8) axisno = comd -8;
            else          axisno = comd;
            if(axisno > 5)  continue;               // ԍŝƂ

            if(comd < 8){                           // xݒ̂Ƃ
                if(cdata == 0xFFFFFFFF) cdata = movevelo[axisno];  // ŵƂ
                if(cdata > 4000000)    cdata = 4000000;            // ől(2MHz)̐
                movevelo[axisno] = cdata;           // xۑ
                i = (axisno << 16) | 0x2100000;     // xݒ |[g1f[^
            }
            else{                                   // xݒ̂Ƃ
                if(cdata == 0xFFFFFFFF) cdata = moveacce[axisno];  // ŵƂ
                if(cdata > 0xFFFF00)    cdata = 0xFFFF00;          // ől̐
                moveacce[axisno] = cdata;           // xۑ
                i = (axisno << 16) | 0x2200000;     // xݒ |[g1f[^
            }

            *(uint32_t *)h2p_lw_pout0_addr = cdata; // x܂͉x|[g0ɃZbg
            *(uint32_t *)h2p_lw_pout1_addr = i;     // VR}hR[h|[g1ɃZbg
            comtrig();                              // R}h݃gK
            
            printf("U%c%c%08X%c",recbuf[1],recbuf[2],cdata,0xd);   // 𑗐M
            continue;
        }

        // QR}h ([VRg[ړʓǎ) ̂Ƃ

        if((recbuf[0] == 'Q') && (recbuf[1] == swID)){

            comd  = 0xFFFFFFFF;
            if(hextobin() != 0) continue;           // 16i̕oCi
            if(comd > 7)  continue;                 // ԍŝƂ
            axisno = comd;

            *(uint32_t *)h2p_lw_pout1_addr = ((axisno << 16) | 0x2800000);
                                                    // [VRg[ԓǎR[h|[g1ɃZbg
            comtrig();                              // R}h݃gK
            i = comread(&pstate[axisno], &ppuls[axisno]);   // ǎs
            if(i > 50){
                printf("FPGA_ERR   %c",0xd);        // BUSYOFFƂȂȂƂ̓G[
                continue;
            }
            if(pstate[axisno] & 1) ppuls[axisno] = 0 - ppuls[axisno]; // Ƃ
            if(pstate[axisno] & 0x100) comd = comd + 8;               // ړƂ

            printf("S%c%1X%08X%c",recbuf[1],comd,ppuls[axisno],0xd);  // 𑗐M
            continue;
        }

        // qR}h ([VRg[݈ʒuǎ) ̂Ƃ

        if((recbuf[0] == 'q') && (recbuf[1] == swID)){

            if((recbuf[2] == 'I') || (recbuf[2] == 'i')){              // ݈ʒuSZbg
                *(uint32_t *)h2p_lw_pout1_addr = 0x2903F00;
                                                    // [VRg[݈ʒuZbgR[h|[g1ɃZbg
                comtrig();                          // R}h݃gK
                comd = 0;
            }
            else{
                comd  = 0xFFFFFFFF;
                cdata = poffset[comd];
                if(hextobin() != 0) continue;       // 16i̕oCi
                if(comd > 7)  continue;             // ԍŝƂ
                if((recbuf[3] != 0xd) && (recbuf[3] != 0) && (recbuf[3] != 0xa)){
                                                    // lw܂݈͌ʒuʃZbĝƂ
                    if(comd < 6){
                        if((recbuf[3] == 'I') || (recbuf[3] == 'i')){  // ݈ʒuʂɃZbg
                            *(uint32_t *)h2p_lw_pout1_addr = ((0x100 << comd) | 0x2900000);
                            comtrig();              // R}h݃gK
                        }
                        else{
                            if((recbuf[3] != 'R') && (recbuf[3] != 'r')){    // l݂̂Ƃ
                                poffset[comd] = cdata; // lۑ
                                *(uint32_t *)h2p_lw_pout0_addr = cdata;      // l|[g0ɃZbg
                                *(uint32_t *)h2p_lw_pout1_addr = ((comd << 16) | 0x2400000);
                                                    // [VRg[l݃R[h|[g1ɃZbg
                                comtrig();          // R}h݃gK
                            }
                        }
                        printf("U%c%1X%08X%c",recbuf[1],comd,poffset[comd],0xd); // l𑗐M
                        continue;
                    }
                }
            }
            axisno = comd;

            *(uint32_t *)h2p_lw_pout1_addr = ((axisno << 16) | 0x2900000);
                                                    // [VRg[݈ʒuǎR[h|[g1ɃZbg
            comtrig();                              // R}h݃gK
            i = comread(&pstate[axisno], &pposi[axisno]);   // ǎs
            if(i > 50){
                printf("FPGA_ERR   %c",0xd);        // BUSYOFFƂȂȂƂ̓G[
                continue;
            }
            if(pstate[axisno] & 0x100) comd = comd + 8;  // ړƂ

            printf("s%c%1X%08X%c",recbuf[1],comd,pposi[axisno],0xd); // ݈ʒu𑗐M
            continue;
        }

        // SR}h (ADCX^[g/Xgbv) ̂Ƃ

        if((recbuf[0] == 'S') && (recbuf[1] == swID)){
            comd  = 0xFFFFFFFF;
            cdata = 0xFFFFFFFF;
            if(hextobin() != 0) continue;        // 16i̕oCi
            if(comd == 13){
                adcsd = 1;                       // ͂Ƃ
                printf("U%c%1X%08X%c",recbuf[1],comd,adcinterval,0xd);   // 𑗐M
                continue;
            }
            if(comd == 14){
                adcsd = 0;                       // VO͂Ƃ
                printf("U%c%1X%08X%c",recbuf[1],comd,adcinterval,0xd);   // 𑗐M
                continue;
            }
            if(comd > 9) continue;               // ԍŝƂ
            if(comd == 9){
                p = 0x3000000;                   // ADC XgbvR[h
                adcstart = 0;
            }
            else{
                p = 0x3000100;                   // ADC X^[gR[h
                adcstart = 1;
            }
            if(cdata != 0xFFFFFFFF) adcinterval = cdata;  // ADC C^[oXV
            if(adcinterval < 10000) adcinterval = 10000;  // ADC C^[oŏl
            if(comd == 8) q = 8;                 // `lXVZbg
            else{
                if(comd == 9) q = adcmode;
                else          q = comd & 7;      // Œ`lZbg
            }
            adcmode = q;                         // ADC 샂[hۑ
            p = p | q;
            if(adcsd) p = p | 0x10;              // ̓Zbg
            
            *(uint32_t *)h2p_lw_pout0_addr = adcinterval;  // ADC C^[o|[g0ɃZbg
            *(uint32_t *)h2p_lw_pout1_addr = p;  // ADCJn/~R[h|[g1ɃZbg
            comtrig();                           // R}h݃gK

            printf("U%c%1X%08X%c",recbuf[1],comd,adcinterval,0xd);   // 𑗐M
            continue;
        }

        // pR}h (PWMo͏) ̂Ƃ

        if((recbuf[0] == 'p') && (recbuf[1] == swID)){
            if((recbuf[2] == 'R') || (recbuf[2] == 'r')){    // PWM moving state ǎŵƂ
                *(uint32_t *)h2p_lw_pout1_addr =  0x4100000; // ǎR[h|[g1ɃZbg
                comtrig();                                   // R}h݃gK
                i = comread(&comd, &cdata);                  // ǎs
                printf("p%c0%08X%c",recbuf[1],cdata,0xd);
                continue;
            }
            cdata = 0;
            comd  = 0xFFFFFFFF;
            if(hextobin() != 0) continue;           // 16i̕oCi
            if(comd >= 0x10) continue;              // `lԍŵƂ

            if(cdata == 0) cdata = pwmwidth[comd];  // PWMpX܂̓TCNŵƂ
            if(comd == 14){                         // PWMJnR}ĥƂ
                hextobin();                         // 0̂Ƃ͋f[^𖳎wlɂčēxoCiϊ
                                                    // S`l~̂߂0w
                cdata = cdata & 0x10003FFF;         // start/stop ̂Ƃ bit31̓pXݒ͈͉
            }
            else{
                if(comd == 15){                     // PWMTCN̍ől
                    if(cdata > 0xFFFFF) cdata = 0xFFFFF;
                }
                else{                               // PWMpX̍ől
                    u = cdata & 0xFFFF;
                    if((pwmwidth[14] & 0x10000000) == 0){  // ̂Ƃ
                        if(u > 62000) u = 62000;    // 2480us
                        if(u < 14000) u = 14000;    // 560us
                    }
                    cdata = (cdata & 0xFFFF0000) | u;
                }
            }
            pwmwidth[comd] = cdata;                 // PWMpX/ωx܂̓TCNZbg
            u = cdata;
            if(comd == 14) u = u & 0x3FFF;
            else{
                if(comd == 15){
                    if(u != 0) u = u - 1;           // TCN+1ƂȂ̂␳
                }
                else{
                    if((u & 0xffff) != 0) u = u - 1;               // pX+1ƂȂ̂␳
                    if((u & 0xffff0000) > 0x10000) u = u -0x10000; // ωx+1ƂȂ̂␳
                }
            }
            *(uint32_t *)h2p_lw_pout0_addr = u;     // PWMpX܂̓TCNf[^|[g0ɃZbg
            *(uint32_t *)h2p_lw_pout1_addr = ((comd << 16) | 0x4000000);  // pR}hR[h|[g1ɃZbg
            comtrig();                              // R}h݃gK
            
            printf("U%c%1X%08X%c",recbuf[1],comd,cdata,0xd);  // 𑗐M
            continue;
        }

        // R}hR[hG[̂Ƃ
        
        printf("U0RX_C%c_ERR%c",recbuf[0],0xd); // R}hR[hG[𑗐M

    }while(1);
    
    fcntl(STDIN_FILENO, F_SETFL, oldf);          // Wo͂ ɖ߂
    
    // FPGA Device/Memory CLOSE
    
    if( munmap( virtual_base, HW_REGS_SPAN ) != 0 ) {
        printf( "MCLR_ERR   %c",0xd);
        close( fd );
        return( 1 );
    }
    close( fd );

    return( 0 );
}

int hextobin(){

    int i, ret;
    unsigned int j, k, temp;

    ret = -1;
    j = 0x20000000;                     // 16iȅZl
    k = 0xf0000000;                     // 16i4bitƂ̃}XN

    for(i=2; i!= 30; i++){
        temp = 0x10;                                           // ܂Af[^Zbg
        if((recbuf[i] == 0xd) || (recbuf[i] == 0) || (recbuf[i] == 0xa)) break;
                                                               // IR[ĥƂ
        if((recbuf[i] == ' ') || (recbuf[i] == '_')) continue; // Xy[XƃA_[o[͓ǔ΂
        if((recbuf[i] >= '0') && (recbuf[i] <= '9')) temp = (int)(recbuf[i]-0x30);         // 0`9
        else{
            if((recbuf[i] >= 'A') && (recbuf[i] <= 'F')) temp = (int)(recbuf[i]-0x37);     // A`F
            else{
                if((recbuf[i] >= 'a') && (recbuf[i] <= 'f')) temp = (int)(recbuf[i]-0x57); // a`f
            }
        }
        if(j == 0x20000000){                                   // ŏʌ(bit35`32)̂Ƃ
            if(temp != 0x10) comd = temp;                      // go̓f[^ɃZbg
            j = 0x10000000;                                    // 16iZlZbg
        }
        else{
            if(temp != 0x10) cdata = (cdata & ~k) | (j*temp);  // Y̏o͂Zbg
            j = j >> 4;                                        // 16iZlZbg
            if(j == 0) break;                                  // ŏI(bit3`0)IƂ
            k = (k >> 4) & 0x0fffffff;                         // 16i4bit}XNZbg
        }
    }
    if(i < 30) ret = 0;

    return(ret);
}

void comtrig(){
    *(uint32_t *)h2p_lw_pout2_addr = comport2on;               // R}h݃gKON
    *(uint32_t *)h2p_lw_pout2_addr = comport2off;              // R}h݃gKOFF
    return;
}

int comread(unsigned int *d, unsigned int *c){
    int i;

    for(i=0; i!= 100; i++){
        if((*(uint32_t *)h2p_lw_pin2_addr & 0xc0000000) == 0){ // BUSYOFF̂Ƃ
            *c = *(uint32_t *)h2p_lw_pin0_addr;                // |[g0ǎ
            *d = *(uint32_t *)h2p_lw_pin1_addr;                // |[g1ǎ
            break;
        }
    }
    return(i);
}
